"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.onPostConfigureApp = exports.onDownloadApp = exports.buildSafariPreferences = exports.isolateAppBundle = exports.unzipStream = exports.unzipFile = exports.isAppBundle = exports.parseLocalizableStrings = exports.verifyApplicationPlatform = exports.fetchSupportedAppPlatforms = exports.extractBundleVersion = exports.extractBundleId = exports.SUPPORTED_EXTENSIONS = exports.IPA_EXT = exports.APP_EXT = exports.SAFARI_BUNDLE_ID = void 0;
const lodash_1 = __importDefault(require("lodash"));
const path_1 = __importDefault(require("path"));
const support_1 = require("appium/support");
const logger_js_1 = __importDefault(require("./logger.js"));
const lru_cache_1 = require("lru-cache");
const node_os_1 = __importDefault(require("node:os"));
const teen_process_1 = require("teen_process");
const bluebird_1 = __importDefault(require("bluebird"));
const node_child_process_1 = require("node:child_process");
const node_assert_1 = __importDefault(require("node:assert"));
const STRINGSDICT_RESOURCE = '.stringsdict';
const STRINGS_RESOURCE = '.strings';
exports.SAFARI_BUNDLE_ID = 'com.apple.mobilesafari';
exports.APP_EXT = '.app';
exports.IPA_EXT = '.ipa';
/** @type {LRUCache<string, import('@appium/types').StringRecord>} */
const PLIST_CACHE = new lru_cache_1.LRUCache({ max: 20 });
const SAFARI_OPTS_ALIASES_MAP = /** @type {const} */ ({
    safariAllowPopups: [
        ['WebKitJavaScriptCanOpenWindowsAutomatically', 'JavaScriptCanOpenWindowsAutomatically'],
        (x) => Number(Boolean(x)),
    ],
    safariIgnoreFraudWarning: [['WarnAboutFraudulentWebsites'], (x) => Number(!x)],
    safariOpenLinksInBackground: [['OpenLinksInBackground'], (x) => Number(Boolean(x))],
});
const MAX_ARCHIVE_SCAN_DEPTH = 1;
exports.SUPPORTED_EXTENSIONS = [exports.IPA_EXT, exports.APP_EXT];
const MACOS_RESOURCE_FOLDER = '__MACOSX';
/**
 * Retrieves the value of the given entry name from the application's Info.plist.
 *
 * @this {Object} Optinal instance used for caching. Ususally the driver instance.
 * @param {string} app Full path to the app bundle root.
 * @param {string} entryName Key name in the plist.
 * @returns {Promise<any | undefined>} Either the extracted value or undefined if no such key has been found in the plist.
 * @throws {Error} If the application's Info.plist cannot be parsed.
 */
async function extractPlistEntry(app, entryName) {
    const plistPath = path_1.default.resolve(app, 'Info.plist');
    const parseFile = async () => {
        try {
            return await support_1.plist.parsePlistFile(plistPath);
        }
        catch (err) {
            throw new Error(`Could not extract Info.plist from '${path_1.default.basename(app)}': ${err.message}`);
        }
    };
    let plistObj = PLIST_CACHE.get(app);
    if (!plistObj) {
        plistObj = await parseFile();
        PLIST_CACHE.set(app, plistObj);
    }
    return /** @type {import('@appium/types').StringRecord} */ (plistObj)[entryName];
}
/**
 *
 * @param {string} app
 * @returns {Promise<string>}
 */
async function extractBundleId(app) {
    const bundleId = await extractPlistEntry(app, 'CFBundleIdentifier');
    logger_js_1.default.debug(`Getting bundle ID from app '${app}': '${bundleId}'`);
    return bundleId;
}
exports.extractBundleId = extractBundleId;
/**
 *
 * @param {string} app
 * @returns {Promise<string>}
 */
async function extractBundleVersion(app) {
    return await extractPlistEntry(app, 'CFBundleVersion');
}
exports.extractBundleVersion = extractBundleVersion;
/**
 *
 * @param {string} app
 * @returns {Promise<string>}
 */
async function extractExecutableName(app) {
    return await extractPlistEntry(app, 'CFBundleExecutable');
}
/**
 *
 * @param {string} app
 * @returns {Promise<string[]>}
 */
async function fetchSupportedAppPlatforms(app) {
    try {
        const result = await extractPlistEntry(app, 'CFBundleSupportedPlatforms');
        if (!lodash_1.default.isArray(result)) {
            logger_js_1.default.warn(`${path_1.default.basename(app)}': CFBundleSupportedPlatforms is not a valid list`);
            return [];
        }
        return result;
    }
    catch (err) {
        logger_js_1.default.warn(`Cannot extract the list of supported platforms from '${path_1.default.basename(app)}': ${err.message}`);
        return [];
    }
}
exports.fetchSupportedAppPlatforms = fetchSupportedAppPlatforms;
/**
 * @typedef {Object} PlatformOpts
 *
 * @property {boolean} isSimulator - Whether the destination platform is a Simulator
 * @property {boolean} isTvOS - Whether the destination platform is a Simulator
 */
/**
 * Verify whether the given application is compatible to the
 * platform where it is going to be installed and tested.
 *
 * @param {string} app - The actual path to the application bundle
 * @param {PlatformOpts} expectedPlatform
 * @throws {Error} If bundle architecture does not match the expected device architecture.
 */
async function verifyApplicationPlatform(app, expectedPlatform) {
    logger_js_1.default.debug('Verifying application platform');
    const supportedPlatforms = await fetchSupportedAppPlatforms(app);
    logger_js_1.default.debug(`CFBundleSupportedPlatforms: ${JSON.stringify(supportedPlatforms)}`);
    const { isSimulator, isTvOS } = expectedPlatform;
    const prefix = isTvOS ? 'AppleTV' : 'iPhone';
    const suffix = isSimulator ? 'Simulator' : 'OS';
    const dstPlatform = `${prefix}${suffix}`;
    const appFileName = path_1.default.basename(app);
    if (!supportedPlatforms.includes(dstPlatform)) {
        throw new Error(`${isSimulator ? 'Simulator' : 'Real device'} architecture is not supported by the '${appFileName}' application. ` +
            `Make sure the correct deployment target has been selected for its compilation in Xcode.`);
    }
    if (isSimulator) {
        const executablePath = path_1.default.resolve(app, await extractExecutableName(app));
        const [resFile, resUname] = await bluebird_1.default.all([
            (0, teen_process_1.exec)('file', [executablePath]),
            (0, teen_process_1.exec)('uname', ['-m']),
        ]);
        const bundleExecutableInfo = lodash_1.default.trim(resFile.stdout);
        logger_js_1.default.debug(bundleExecutableInfo);
        const arch = lodash_1.default.trim(resUname.stdout);
        const isAppleSilicon = node_os_1.default.cpus()[0].model.includes('Apple');
        // We cannot run Simulator builds compiled for arm64 on Intel machines
        // Rosetta allows only to run Intel ones on arm64
        if (!lodash_1.default.includes(bundleExecutableInfo, `executable ${arch}`) &&
            !(isAppleSilicon && lodash_1.default.includes(bundleExecutableInfo, 'executable x86_64'))) {
            const bundleId = await extractBundleId(app);
            throw new Error(`The ${bundleId} application does not support the ${arch} Simulator ` +
                `architecture:\n${bundleExecutableInfo}\n\n` +
                `Please rebuild your application to support the ${arch} platform.`);
        }
    }
}
exports.verifyApplicationPlatform = verifyApplicationPlatform;
async function readResource(resourcePath) {
    const data = await support_1.plist.parsePlistFile(resourcePath);
    const result = {};
    for (const [key, value] of lodash_1.default.toPairs(data)) {
        result[key] = lodash_1.default.isString(value) ? value : JSON.stringify(value);
    }
    return result;
}
async function parseLocalizableStrings(opts) {
    const { app, language = 'en', localizableStringsDir, stringFile, strictMode } = opts;
    if (!app) {
        const message = `Strings extraction is not supported if 'app' capability is not set`;
        if (strictMode) {
            throw new Error(message);
        }
        logger_js_1.default.info(message);
        return {};
    }
    let lprojRoot;
    for (const subfolder of [`${language}.lproj`, localizableStringsDir, '']) {
        lprojRoot = path_1.default.resolve(app, subfolder);
        if (await support_1.fs.exists(lprojRoot)) {
            break;
        }
        const message = `No '${lprojRoot}' resources folder has been found`;
        if (strictMode) {
            throw new Error(message);
        }
        logger_js_1.default.debug(message);
    }
    logger_js_1.default.info(`Will extract resource strings from '${lprojRoot}'`);
    const resourcePaths = [];
    if (stringFile) {
        const dstPath = path_1.default.resolve(String(lprojRoot), stringFile);
        if (await support_1.fs.exists(dstPath)) {
            resourcePaths.push(dstPath);
        }
        else {
            const message = `No '${dstPath}' resource file has been found for '${app}'`;
            if (strictMode) {
                throw new Error(message);
            }
            logger_js_1.default.info(message);
            logger_js_1.default.info(`Getting all the available strings from '${lprojRoot}'`);
        }
    }
    if (lodash_1.default.isEmpty(resourcePaths) && (await support_1.fs.exists(String(lprojRoot)))) {
        const resourceFiles = (await support_1.fs.readdir(String(lprojRoot)))
            .filter((name) => lodash_1.default.some([STRINGS_RESOURCE, STRINGSDICT_RESOURCE], (x) => name.endsWith(x)))
            .map((name) => path_1.default.resolve(lprojRoot, name));
        resourcePaths.push(...resourceFiles);
    }
    logger_js_1.default.info(`Got ${resourcePaths.length} resource file(s) in '${lprojRoot}'`);
    if (lodash_1.default.isEmpty(resourcePaths)) {
        return {};
    }
    const resultStrings = {};
    const toAbsolutePath = function (p) {
        return path_1.default.isAbsolute(p) ? p : path_1.default.resolve(process.cwd(), p);
    };
    for (const resourcePath of resourcePaths) {
        if (!support_1.util.isSubPath(toAbsolutePath(resourcePath), toAbsolutePath(app))) {
            // security precaution
            throw new Error(`'${resourcePath}' is expected to be located under '${app}'`);
        }
        try {
            const data = await readResource(resourcePath);
            logger_js_1.default.debug(`Parsed ${lodash_1.default.keys(data).length} string(s) from '${resourcePath}'`);
            lodash_1.default.merge(resultStrings, data);
        }
        catch (e) {
            logger_js_1.default.warn(`Cannot parse '${resourcePath}' resource. Original error: ${e.message}`);
        }
    }
    logger_js_1.default.info(`Got ${lodash_1.default.keys(resultStrings).length} string(s) from '${lprojRoot}'`);
    return resultStrings;
}
exports.parseLocalizableStrings = parseLocalizableStrings;
/**
 * Check whether the given path on the file system points to the .app bundle root
 *
 * @param {string} appPath Possible .app bundle root
 * @returns {Promise<boolean>} Whether the given path points to an .app bundle
 */
async function isAppBundle(appPath) {
    return (lodash_1.default.endsWith(lodash_1.default.toLower(appPath), exports.APP_EXT) &&
        (await support_1.fs.stat(appPath)).isDirectory() &&
        (await support_1.fs.exists(path_1.default.join(appPath, 'Info.plist'))));
}
exports.isAppBundle = isAppBundle;
/**
 * @typedef {Object} UnzipInfo
 * @property {string} rootDir
 * @property {number} archiveSize
 */
/**
 * Unzips a ZIP archive on the local file system.
 *
 * @param {string} archivePath Full path to a .zip archive
 * @returns {Promise<UnzipInfo>} temporary folder root where the archive has been extracted
 */
async function unzipFile(archivePath) {
    const useSystemUnzipEnv = process.env.APPIUM_PREFER_SYSTEM_UNZIP;
    const useSystemUnzip = lodash_1.default.isEmpty(useSystemUnzipEnv) || !['0', 'false'].includes(lodash_1.default.toLower(useSystemUnzipEnv));
    const tmpRoot = await support_1.tempDir.openDir();
    try {
        await support_1.zip.extractAllTo(archivePath, tmpRoot, { useSystemUnzip });
    }
    catch (e) {
        await support_1.fs.rimraf(tmpRoot);
        throw e;
    }
    return {
        rootDir: tmpRoot,
        archiveSize: (await support_1.fs.stat(archivePath)).size,
    };
}
exports.unzipFile = unzipFile;
/**
 * Unzips a ZIP archive from a stream.
 * Uses bdstar tool for this purpose.
 * This allows to optimize the time needed to prepare the app under test
 * to MAX(download, unzip) instead of SUM(download, unzip)
 *
 * @param {import('node:stream').Readable} zipStream
 * @returns {Promise<UnzipInfo>}
 */
async function unzipStream(zipStream) {
    const tmpRoot = await support_1.tempDir.openDir();
    const bsdtarProcess = (0, node_child_process_1.spawn)(await support_1.fs.which('bsdtar'), [
        '-x',
        '--exclude', MACOS_RESOURCE_FOLDER,
        '--exclude', `${MACOS_RESOURCE_FOLDER}/*`,
        '-',
    ], {
        cwd: tmpRoot,
    });
    let archiveSize = 0;
    bsdtarProcess.stderr.on('data', (chunk) => {
        const stderr = chunk.toString();
        if (lodash_1.default.trim(stderr)) {
            logger_js_1.default.warn(stderr);
        }
    });
    zipStream.on('data', (chunk) => {
        archiveSize += lodash_1.default.size(chunk);
    });
    zipStream.pipe(bsdtarProcess.stdin);
    try {
        await new bluebird_1.default((resolve, reject) => {
            zipStream.once('error', reject);
            bsdtarProcess.once('exit', (code, signal) => {
                zipStream.unpipe(bsdtarProcess.stdin);
                logger_js_1.default.debug(`bsdtar process exited with code ${code}, signal ${signal}`);
                if (code === 0) {
                    resolve();
                }
                else {
                    reject(new Error('Is it a valid ZIP archive?'));
                }
            });
            bsdtarProcess.once('error', (e) => {
                zipStream.unpipe(bsdtarProcess.stdin);
                reject(e);
            });
        });
    }
    catch (err) {
        bsdtarProcess.kill(9);
        await support_1.fs.rimraf(tmpRoot);
        throw new Error(`The response data cannot be unzipped: ${err.message}`);
    }
    finally {
        bsdtarProcess.removeAllListeners();
        zipStream.removeAllListeners();
    }
    return {
        rootDir: tmpRoot,
        archiveSize,
    };
}
exports.unzipStream = unzipStream;
/**
 * Looks for items with given extensions in the given folder
 *
 * @param {string} appPath Full path to an app bundle
 * @param {Array<string>} appExtensions List of matching item extensions
 * @returns {Promise<string[]>} List of relative paths to matched items
 */
async function findApps(appPath, appExtensions) {
    const globPattern = `**/*.+(${appExtensions.map((ext) => ext.replace(/^\./, '')).join('|')})`;
    const sortedBundleItems = (await support_1.fs.glob(globPattern, {
        cwd: appPath,
    })).sort((a, b) => a.split(path_1.default.sep).length - b.split(path_1.default.sep).length);
    return sortedBundleItems;
}
/**
 * Moves the application bundle to a newly created temporary folder
 *
 * @param {string} appRoot Full path to the .app bundle
 * @returns {Promise<string>} The new path to the app bundle.
 * The name of the app bundle remains though
 */
async function isolateAppBundle(appRoot) {
    const tmpRoot = await support_1.tempDir.openDir();
    const dstRoot = path_1.default.join(tmpRoot, path_1.default.basename(appRoot));
    await support_1.fs.mv(appRoot, dstRoot, { mkdirp: true });
    return dstRoot;
}
exports.isolateAppBundle = isolateAppBundle;
/**
 * Builds Safari preferences object based on the given session capabilities
 *
 * @param {import('./driver').XCUITestDriverOpts} opts
 * @return {Promise<import('@appium/types').StringRecord>}
 */
function buildSafariPreferences(opts) {
    const safariSettings = lodash_1.default.cloneDeep(opts?.safariGlobalPreferences ?? {});
    for (const [name, [aliases, valueConverter]] of lodash_1.default.toPairs(SAFARI_OPTS_ALIASES_MAP)) {
        if (!lodash_1.default.has(opts, name)) {
            continue;
        }
        for (const alias of aliases) {
            safariSettings[alias] = valueConverter(opts[name]);
        }
    }
    return safariSettings;
}
exports.buildSafariPreferences = buildSafariPreferences;
/**
 * Unzip the given archive and find a matching .app bundle in it
 *
 * @this {import('./driver').XCUITestDriver}
 * @param {string|import('node:stream').Readable} appPathOrZipStream The path to the archive.
 * @param {number} depth [0] the current nesting depth. App bundles whose nesting level
 * is greater than 1 are not supported.
 * @returns {Promise<string>} Full path to the first matching .app bundle..
 * @throws If no matching .app bundles were found in the provided archive.
 */
async function unzipApp(appPathOrZipStream, depth = 0) {
    const errMsg = `The archive '${this.opts.app}' did not have any matching ${exports.APP_EXT} or ${exports.IPA_EXT} ` +
        `bundles. Please make sure the provided package is valid and contains at least one matching ` +
        `application bundle which is not nested.`;
    if (depth > MAX_ARCHIVE_SCAN_DEPTH) {
        throw new Error(errMsg);
    }
    const timer = new support_1.timing.Timer().start();
    /** @type {string} */
    let rootDir;
    /** @type {number} */
    let archiveSize;
    try {
        if (lodash_1.default.isString(appPathOrZipStream)) {
            ({ rootDir, archiveSize } = await unzipFile(appPathOrZipStream));
        }
        else {
            if (depth > 0) {
                node_assert_1.default.fail('Streaming unzip cannot be invoked for nested archive items');
            }
            ({ rootDir, archiveSize } = await unzipStream(appPathOrZipStream));
        }
    }
    catch (e) {
        this.log.debug(e.stack);
        throw new Error(`Cannot prepare the application at '${this.opts.app}' for testing. Original error: ${e.message}`);
    }
    const secondsElapsed = timer.getDuration().asSeconds;
    this.log.info(`The app '${this.opts.app}' (${support_1.util.toReadableSizeString(archiveSize)}) ` +
        `has been ${lodash_1.default.isString(appPathOrZipStream) ? 'extracted' : 'downloaded and extracted'} ` +
        `to '${rootDir}' in ${secondsElapsed.toFixed(3)}s`);
    // it does not make much sense to approximate the speed for short downloads
    if (secondsElapsed >= 1) {
        const bytesPerSec = Math.floor(archiveSize / secondsElapsed);
        this.log.debug(`Approximate decompression speed: ${support_1.util.toReadableSizeString(bytesPerSec)}/s`);
    }
    const matchedPaths = await findApps(rootDir, exports.SUPPORTED_EXTENSIONS);
    if (lodash_1.default.isEmpty(matchedPaths)) {
        this.log.debug(`'${path_1.default.basename(rootDir)}' has no bundles`);
    }
    else {
        this.log.debug(`Found ${support_1.util.pluralize('bundle', matchedPaths.length, true)} in ` +
            `'${path_1.default.basename(rootDir)}': ${matchedPaths}`);
    }
    try {
        for (const matchedPath of matchedPaths) {
            const fullPath = path_1.default.join(rootDir, matchedPath);
            if (await isAppBundle(fullPath)) {
                const supportedPlatforms = await fetchSupportedAppPlatforms(fullPath);
                if (this.isSimulator() && !supportedPlatforms.some((p) => lodash_1.default.includes(p, 'Simulator'))) {
                    this.log.info(`'${matchedPath}' does not have Simulator devices in the list of supported platforms ` +
                        `(${supportedPlatforms.join(',')}). Skipping it`);
                    continue;
                }
                if (this.isRealDevice() && !supportedPlatforms.some((p) => lodash_1.default.includes(p, 'OS'))) {
                    this.log.info(`'${matchedPath}' does not have real devices in the list of supported platforms ` +
                        `(${supportedPlatforms.join(',')}). Skipping it`);
                    continue;
                }
                this.log.info(`'${matchedPath}' is the resulting application bundle selected from '${rootDir}'`);
                return await isolateAppBundle(fullPath);
            }
            else if (lodash_1.default.endsWith(lodash_1.default.toLower(fullPath), exports.IPA_EXT) && (await support_1.fs.stat(fullPath)).isFile()) {
                try {
                    return await unzipApp.bind(this)(fullPath, depth + 1);
                }
                catch (e) {
                    this.log.warn(`Skipping processing of '${matchedPath}': ${e.message}`);
                }
            }
        }
    }
    finally {
        await support_1.fs.rimraf(rootDir);
    }
    throw new Error(errMsg);
}
/**
 * @this {import('./driver').XCUITestDriver}
 * @param {import('@appium/types').DownloadAppOptions} opts
 * @returns {Promise<string>}
 */
async function onDownloadApp({ stream }) {
    return await unzipApp.bind(this)(stream);
}
exports.onDownloadApp = onDownloadApp;
/**
 * @this {import('./driver').XCUITestDriver}
 * @param {import('@appium/types').PostProcessOptions} opts
 * @returns {Promise<import('@appium/types').PostProcessResult|false>}
 */
async function onPostConfigureApp({ cachedAppInfo, isUrl, appPath }) {
    // Pick the previously cached entry if its integrity has been preserved
    /** @type {import('@appium/types').CachedAppInfo|undefined} */
    const appInfo = lodash_1.default.isPlainObject(cachedAppInfo) ? cachedAppInfo : undefined;
    const cachedPath = appInfo ? /** @type {string} */ (appInfo.fullPath) : undefined;
    if (
    // If cache is present
    appInfo && cachedPath
        // And if the path exists
        && await support_1.fs.exists(cachedPath)
        // And if hash matches to the cached one if this is a file
        // Or count of files >= of the cached one if this is a folder
        && (((await support_1.fs.stat(cachedPath)).isFile()
            && await support_1.fs.hash(cachedPath) === /** @type {any} */ (appInfo.integrity)?.file)
            || (await support_1.fs.glob('**/*', { cwd: cachedPath })).length >= /** @type {any} */ (appInfo.integrity)?.folder)) {
        this.log.info(`Using '${cachedPath}' which was cached from '${appPath}'`);
        return { appPath: cachedPath };
    }
    const isBundleAlreadyUnpacked = await isAppBundle(/** @type {string} */ (appPath));
    // Only local .app bundles that are available in-place should not be cached
    if (!isUrl && isBundleAlreadyUnpacked) {
        return false;
    }
    // Cache the app while unpacking the bundle if necessary
    return {
        appPath: isBundleAlreadyUnpacked ? appPath : await unzipApp.bind(this)(/** @type {string} */ (appPath))
    };
}
exports.onPostConfigureApp = onPostConfigureApp;
//# sourceMappingURL=app-utils.js.map